#pragma once

#include "volumeshop.h"

#include "ImporterDicom.h"
#include "Volume.h"

template <template <class,unsigned int> class ResourceType, class ElementType, unsigned int N, class ParameterType>
class LoaderDicom : public Loader
{
public:
	LoaderDicom(ResourceType<ElementType,N> & resResource, const ParameterType & parParameter)
	{
	};
};

template<class ResourceType, unsigned int Components>
class LoaderDicom<Volume,ResourceType,Components,ImporterDicom> : public Loader, public Observer
{
public:
	LoaderDicom(Volume<ResourceType,Components> & volVolume, const ImporterDicom & impImporter) : m_volVolume(volVolume)
	{
		m_modVolumeObserver.connect(this,&LoaderDicom::volumeDestroyed);
		volVolume.addObserver(&m_modVolumeObserver);

		// init series
		_serieHelper = gdcm::SerieHelper::New();

		// get all filenames
		const Variant & filename = impImporter.GetPlugin().GetProperty("Filename");
		for (int i=0; i<filename.GetSize(); i++)
			_serieHelper->AddFileName(std::string(filename[i]));

		// set which serie to load
		_fileList = _serieHelper->GetFirstSingleSerieUIDFileSet();
		if (impImporter.GetPlugin().HasProperty("Series")) {
			try {
				const Variant series = impImporter.GetPlugin().GetProperty("Series");
				for (int i=0; i<series.GetSize(); i++) {
					if (series[i] == std::string(series))
						break;
					_fileList = _serieHelper->GetNextSingleSerieUIDFileSet();
				}
			} catch (Exception e) {}
		}

		// sort the images.
		// returns "position is not unique" on some DICOMs.
		_serieHelper->OrderFileList(_fileList);

		// import
		elastographyMode = impImporter.GetPlugin().GetProperty("Elastography");

		// 
		slice = impImporter.GetPlugin().GetProperty("Slice");

		// export
		m_volVolume.GetProperty("Filename") = filename[0];
		m_volVolume.GetProperty("Elastrography") = elastographyMode;

		_position = 0;
		_width = 0;
		_height = 0;
		_depth = 0;
		_firstPosition = true;
		m_bCancelled = false;
	};

	virtual ~LoaderDicom()
		{
			if (!isCancelled())
			{
				m_volVolume.swap(m_volImport);
				m_volVolume.copyProperties(m_volImport);
			}

			// delete series
			_serieHelper->Delete();
		};

		virtual const std::string GetMessage() const
		{
			return m_strMessage;
		};

		// progress bar
		virtual const float GetProgress() const
		{
			return float(_position) / float(_fileList->size());
		};

		virtual const bool IsAtEnd() const
		{
			if (m_bCancelled)
				return true;
			return _position >= _fileList->size();
		};

		// is cancelled
		virtual const bool isCancelled() const {
			return m_bCancelled;
		}

		// load a slice
		virtual void load()
		{
			if (!IsAtEnd())
			{
				gdcm::File *file = (*_fileList)[_position];
				if (file->IsReadable())
				{
					// load file
					file->Load();
					gdcm::FileHelper *fileHelper = gdcm::FileHelper::New(file);

					// get dicom data
					std::string filename = file->GetFileName();
					std::string sliceThickness = file->GetEntryString(0x0018, 0x0050);
					std::string spaceBetweenSlices = file->GetEntryString(0x0018, 0x0088);
					std::string imagePositionPatient = file->GetEntryString(0x0020, 0x0032);
					std::string sliceLocation = file->GetEntryString(0x0020, 0x1041);
					int width = file->GetXSize();
					int height = file->GetYSize();
					int depth = file->GetZSize();
					float spacingX = file->GetXSpacing();
					float spacingY = file->GetYSpacing();
					float spacingZ = file->GetZSpacing();
					float rescaleSlope = file->GetRescaleSlope();
					float rescaleIntercept = file->GetRescaleIntercept();
					int bitsStored = file->GetBitsStored();
					int samplesPerPixel = file->GetSamplesPerPixel();
					int bitsAllocated = file->GetBitsAllocated();
					std::string pixelType = file->GetPixelType();
					unsigned char *imageData = fileHelper->GetImageData();
					int size = fileHelper->GetImageDataSize();
					
					int bytesPerComponent = bitsAllocated / 8 * samplesPerPixel;

					// region to show
					int regionX1, regionY1, regionX2, regionY2;
					if (elastographyMode == "Elastography") { // left/elastography
						getTest(imageData, width, height, depth, bytesPerComponent, &regionX1, &regionY1, &regionX2, &regionY2);
					} else if (elastographyMode == "B-Mode") { // right/b-mode
						regionX1 = 380;
						regionY1 = 90;
						regionX2 = 710;
						regionY2 = 480;
					} else {
						regionX1 = regionY1 = 0;
						regionX2 = width; //std::min((unsigned int)400, uWidth);
						regionY2 = height; //std::min((unsigned int)400, uHeight);
					}

					if (_firstPosition) {
						// debug
						std::cout << "NOTE: checking first slice..." << std::endl
							<< "filename:" << filename << std::endl
							<< "size:" << width << "," << height << "," << depth << std::endl
							<< "space:" << spacingX << "," << spacingY << "," << spacingZ << std::endl
							<< "slope/intercept:" << rescaleSlope << "," << rescaleIntercept << std::endl
							<< "bitsStored:" << bitsStored << std::endl
							<< "samplesPerPixel(channels):" << samplesPerPixel << std::endl
							<< "bitsAllocated:" << bitsAllocated << std::endl
							<< "imagDataSize:" << size << std::endl
							<< "pixelType:" << pixelType << std::endl
							<< std::endl;

						Matrix matTransformation;
						matTransformation.scale(Vector(spacingX,spacingY,spacingZ));
						_firstPosition = false;

						// set object size
						int resizeDepth = _depth + depth + (_fileList->size() - _position - 1);
						m_volImport.resize(regionX2 - regionX1, regionY2 - regionY1, resizeDepth);
					}

					retrieve(imageData,width,height,depth,bytesPerComponent, bitsAllocated,
						regionX1, regionY1, regionX2, regionY2,
						pixelType, samplesPerPixel, rescaleSlope, rescaleIntercept);

					_width = std::max(width, _width);
					_height = std::max(height, _height);
					_depth += depth;

					if (fileHelper)
						fileHelper->Delete();

				}
				_position++;
			}
		};

		// loading cancelled
		virtual void cancel() {
			m_bCancelled = true;
		}

	protected:
		/*
			read a slice
		*/
		void retrieve(const unsigned char *imageData,
			const int width, const int height, const int depth,
			const int bytesPerComponent, const int bitsAllocated,
			const int x1, const int y1, const int x2, const int y2,
			const std::string pixelType, const int samplesPerPixel, const float slope, const float intercept)
		{
			Volume<ResourceType,Components>::LinearManipulator m(m_volImport);

			for (unsigned int z=0; z<depth; z++) {
				for (unsigned int y=0; y<y2-y1; y++) {
					for (unsigned int x=0; x<x2-x1; x++) {
						int pos = 0;
						pos += x1 + x;
						pos += (y1 + y) * width;
						pos += z * width * height;
						pos *= bytesPerComponent;

						Voxel<unsigned short, Components> v((unsigned short*)&imageData[pos]);

						Voxel<ResourceType, Components> voxel;
						for (unsigned int i=0; i<Components; i++)
							voxel.Set(i, VoxelTypeConverter<ResourceType>::Get(v.Get(i))*16);

						m.SetPosition(x, y, z + _depth);
						m.Set(voxel);
					}
				}
			}
		};
		virtual void destroyed(const Observable & obsObservable)
		{
			m_bCancelled = true;
		};

/*
	search for colors in a given region and export it
*/
	void getTest(const unsigned char *imageData, const int width, const int height, const int depth, const int bytesPerComponent,
		int *regionX1, int *regionY1, int *regionX2, int *regionY2) {
		int x1 = 45;
		int y1 = 90;
		int x2 = 340;
		int y2 = 480;
		for (int z=0; z<depth; z++) {
			for (int y=y1; y<y2; y++) {
				for (int x=x1; x<x2; x++) {
					int pos = 0;
					pos += x;
					pos += y * width;
					pos += z * width * height;
					pos *= bytesPerComponent;

					Voxel<ResourceType,Components> voxel((ResourceType*)&imageData[pos]);
					bool isColor = voxel.Get(0) != voxel.Get(1) || voxel.Get(0) != voxel.Get(2);

					if (isColor) {
						if (x1 == 45) {
							x1 = x + 5;
							y1 = y = y + 5;
							break;
						}
					} else {
						if (x1 > 45 && x2 == 340) {
							x2 = x - 5;
							break;
						}
						if (y1 > 90) {
							y2 = y - 5;
							*regionX1 = x1;
							*regionY1 = y1;
							*regionX2 = x2;
							*regionY2 = y2;
							return;
						}
					}
				}
			}
		}
		*regionX1 = 380;
		*regionY1 = 90;
		*regionX2 = 710;
		*regionY2 = 480;
	};

	virtual void volumeDestroyed(const Volume<ResourceType,Components> & volVolume, const Observable::DestructionEvent & desEvent)
	{
		m_bCancelled = true;
	}

private:
	std::string elastographyMode;
	int slice;

	ModifiedObserver m_modVolumeObserver;
	Volume<ResourceType,Components> & m_volVolume;

	Volume<ResourceType,Components> m_volImport;
	std::string m_strMessage;
	std::vector<std::string> m_vecFilenames;

	gdcm::SerieHelper *_serieHelper;
	gdcm::FileList *_fileList;

	int _position;
	int _width;
	int _height;
	int _depth;

	bool _firstPosition;
	bool m_bCancelled;
};