// TODO: align boxes :)

#include "RendererLabel.h"
#include "vrc_RendererLabel.cpp"

RendererLabel::RendererLabel(Plugin & pluPlugin) : PluginInstance(pluPlugin)
{
	// observers when some plugin attributes changes
	_observerVolume.connect(this, &RendererLabel::Volumechanged);
	_observerDICOM.connect(this, &RendererLabel::DICOMchanged);
	_observerSnap.connect(this, &RendererLabel::snapToVolumeChanged);

	// get&set plugin attributes
	GetPlugin().GetProperty("Projection Transformation").require(Variant::TypeMatrix());
	GetPlugin().GetProperty("Viewing Transformation").require(Variant::TypeMatrix());
	GetPlugin().GetProperty("Volume").require(Variant::TypeHandle());
	GetPlugin().GetProperty("Volume").addObserver(&_observerVolume);
	GetPlugin().GetProperty("Z scale").require(Variant::TypeFloat());
	GetPlugin().GetProperty("Snap to Volume").addObserver(&_observerSnap);
	GetPlugin().GetProperty("Snap to Volume").require(Variant::TypeBoolean());
	GetPlugin().GetProperty("Dictionary").require(Variant::TypeArray());
	
	// text size for text boxes
	fontHeight = .05;

	// set up 4 text boxes
	for (int i=0; i<4; i++) {
		boxes.push_back(box());
		boxes.back().depth = i;
	}
	boxSelected = NULL;

	// default text box colors
	boxes[0].color = 0xffcc66;
	boxes[1].color = 0xffcccc;
	boxes[2].color = 0xccccff;
	boxes[3].color = 0xcccc99;
	boxMargin = .01;

	// default DICOM elements for text boxes
	boxes[0].item.push_back(DICOMItem(0x0010, 0x000A));
	boxes[0].item.push_back(DICOMItem(0x0010, 0x000B));
	boxes[0].item.push_back(DICOMItem(0x0010, 0x000C));
	boxes[0].item.push_back(DICOMItem(0x0010, 0x000D));
	boxes[0].item.push_back(DICOMItem(0x0010, 0x000E));
	boxes[1].item.push_back(DICOMItem(0x0020, 0x000A));
	boxes[1].item.push_back(DICOMItem(0x0020, 0x000B));
	boxes[2].item.push_back(DICOMItem(0x0030, 0x000A));
	boxes[2].item.push_back(DICOMItem(0x0030, 0x000B));
	boxes[3].item.push_back(DICOMItem(0x0040, 0x000A));
	boxes[3].item.push_back(DICOMItem(0x0040, 0x000B));

	_file = gdcm::File::New();
	_updateDICOMProperty = false;

	// align the text boxes
	alignTextBoxes();
}

RendererLabel::~RendererLabel()
{
	if (_file)
		_file->Delete();
}

/*
	Clear DICOM element list when new volume is selected
*/
void RendererLabel::Volumechanged(const Variant & variant, const Observable::Event & event)
{
	GetPlugin().GetProperty("Dictionary").clear();
	GetPlugin().update();
}

void RendererLabel::DICOMchanged(const Variant & variant, const Observable::Event & event)
{
	if (_updateDICOMProperty)
		GetPlugin().update();
}

/*
	set snap-to-volume or free-mode
*/
void RendererLabel::snapToVolumeChanged(const Variant & variant, const Observable::Event & event)
{
	if (!GetPlugin().GetProperty("Snap to Volume"))
		alignTextBoxes();
	GetPlugin().update();
}

/*
	populate DICOM list
*/
void RendererLabel::setDICOMList()
{
	_elementList.clear();
	_file = gdcm::File::New();

	const Handle handleVolume = GetPlugin().GetProperty("Volume").require(Variant::TypeHandle());

	if (!handleVolume.IsValid())
		return;
	// get DICOM header data from first file
	if (_file->IsEmpty()) {
		_file->SetFileName(handleVolume->GetProperty("Filename"));
		_file->Load();
	}

	if (!_file->IsReadable())
		return;

	// get all DICOM data
	gdcm::DocEntry *docEntry = _file->GetFirstEntry();
	_updateDICOMProperty = false;
	for (int i=0; docEntry!=NULL; docEntry=_file->GetNextEntry()) {
		unsigned int group = docEntry->GetKey().GetGroup();
		unsigned int element = docEntry->GetKey().GetElement();
		std::string name = docEntry->GetName();
		std::string value = _file->GetEntryString(group, element);
		
		if (name == "gdcm::Unknown")
			continue;

		GetPlugin().GetProperty("Dictionary")[i].require(Variant::TypeArray());
		GetPlugin().GetProperty("Dictionary")[i][0].require(Variant::TypeString());
		GetPlugin().GetProperty("Dictionary")[i][0] = name;
		GetPlugin().GetProperty("Dictionary")[i][1].require(Variant::TypeString());
//		GetPlugin().GetProperty("Dictionary")[i][1].addObserver(&_observerDICOM);
		GetPlugin().GetProperty("Dictionary")[i][1] = std::string("9");

		_elementList.push_back(gdcm::DictEntry::New(group, element));
		i++;
	}
	_updateDICOMProperty = true;
}

/*
	paint main function
*/
void RendererLabel::display(Canvas & canCanvas)
{
	if (GetPlugin().GetProperty("Dictionary").GetSize() == 0)
		setDICOMList();
	if (GetPlugin().GetProperty("Snap to Volume"))
		snapTextBoxes();
	displayTextBoxes();
}

/*
	paint text boxes
*/
void RendererLabel::displayTextBoxes()
{
	Font font = Font("Arial", 10);

	for (int i=0; i<boxes.size(); i++) {
		struct box *box = &boxes[i];

		// convert 16 bits RGB to 3 components float
		float r = (box->color >> 16) % 256 / 255.0f;
		float g = (box->color >> 8) % 256 / 255.0f;
		float b = box->color % 256 / 255.0f;
		
		glPushMatrix();
		glTranslatef(box->x, box->y, 0);

		// aspect
		int matViewport[4];
		glGetIntegerv(GL_VIEWPORT, matViewport);
		float aspect = (float) matViewport[2] / matViewport[3];
//		glScalef(1, aspect, 1);

		// fill box
		glPolygonMode(GL_BACK, GL_FILL);
		glColor3f(r, g, b);
		glBegin(GL_QUADS);
		glVertex3f(-box->width / 2, box->height /2, 0);
		glVertex3f(box->width / 2, box->height / 2, 0);
		glVertex3f(box->width / 2, -box->height / 2, 0);
		glVertex3f(-box->width / 2, -box->height / 2, 0);
		glEnd();

		// text
		glColor3f(0, 0, 0);
		for (unsigned int i=0; i<box->item.size(); i++) {
			glRasterPos2f(.02f - box->width / 2, -fontHeight * i - fontHeight + box->height / 2);
			font.display("%s", box->item[i].getValue().c_str());
		}

		// line box
		glPolygonMode(GL_BACK, GL_LINE);
		glLineWidth(1);
		if (box == boxSelected) // marks selected box
			glColor3f(.2f, .2f, .2f);
		else
			glColor3f(.8f, .8f, .8f);
		glBegin(GL_QUADS);
		glVertex3f(-box->width / 2, box->height / 2, 0);
		glVertex3f(box->width / 2, box->height / 2, 0);
		glVertex3f(box->width / 2, -box->height / 2, 0);
		glVertex3f(-box->width / 2, -box->height / 2, 0);
		glEnd();

		glPopMatrix();
	}
}

/*
	mouse events
*/
void RendererLabel::button(const Controller & conController)
{
	// stop if auto-move-boxes
	if (GetPlugin().GetProperty("Snap to Volume"))
		return;

	// put each box to each corner
	if (conController.GetButtonState(2) == 1) { // when middle button gets pressed
		alignTextBoxes();
		GetPlugin().update();
		return;
	}

	// moves selected box on top
	if (conController.GetButtonState(0) != 0) {
		const Vector position = conController.GetActiveCursorPosition();
		if (conController.GetButtonState(0) == 1) { // when left button gets pressed
			for (int i=0; i<boxes.size(); i++) {
				int j = boxes.size() - i - 1;
				if (hitTest(position, &boxes[j])) {
					boxes.push_back(boxes[j]);
					boxes.erase(boxes.begin() + j);
					boxSelected = &boxes.back();
					_xOffset = position.GetX() - boxSelected->x;
					_yOffset = position.GetY() - boxSelected->y;
					GetPlugin().update();
					return;
				}
			}
		}
		boxSelected = NULL;
		GetPlugin().update();
	}
}

/*
	mouse movements
*/
void RendererLabel::cursor(const Controller & conController)
{
	// stop if auto-move-boxes
	if (GetPlugin().GetProperty("Snap to Volume"))
		return;

	// move text box
	const Vector position = conController.GetActiveCursorPosition();
	if (boxSelected != NULL) {
		if (conController.GetActiveButtonState()) {
			switch(conController.GetActiveButton()) {
				case 0:
					boxSelected->x = position.GetX() - _xOffset;
					boxSelected->y = position.GetY() - _yOffset;
					GetPlugin().update();
					break;
				case 1:
	//				std::cout << "right" << std::endl;
					break;
				case 2:
	//				std::cout << "center" << std::endl;
					break;
				default:
					break;
			}
		}
	}
}

/*
	put each text box in a corner
*/
void RendererLabel::alignTextBoxes()
{
	float minHeight = .2f;
	for (int i=0; i<boxes.size(); i++) {
		boxes[i].width = .5f;
		boxes[i].height = (boxes[i].item.size() + 1) * fontHeight;
		if (boxes[i].height < minHeight)
			boxes[i].height = minHeight;
		boxes[i].x = boxes[i].depth % 2 ? 1 - boxes[i].width / 2 - boxMargin : boxes[i].width / 2 + boxMargin - 1;
		boxes[i].y = boxes[i].depth / 2 < 1 ? 1 - boxes[i].height / 2 - boxMargin : boxes[i].height / 2 + boxMargin - 1;
	}
}


/*
	snap text box to the volume
*/
void RendererLabel::snapTextBoxes()
{
	const Handle volume = GetPlugin().GetProperty("Volume");
	if (volume.IsValid()) {
		// get volume transformation
		const Matrix projectionTransformation = GetPlugin().GetProperty("Projection Transformation").require(Variant::TypeMatrix());
		const Matrix viewingTransformation = GetPlugin().GetProperty("Viewing Transformation").require(Variant::TypeMatrix());
		const Matrix volumeTransformation = volume->GetProperty("Transformation").require(Variant::TypeMatrix());
		const float dimension = volume->GetSize().GetMaximum();
		const Vector maximum = volume->GetBounds().GetMaximum();
		Matrix dataTransformation;
		dataTransformation *= volumeTransformation;
		dataTransformation.scale(Vector(1 / dimension, 1 / dimension, 1 / dimension));
		dataTransformation.translate(maximum * -0.5f);

		// do matrix multiplications
		Matrix projectionMatrix;
		Matrix modelMatrix;
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadMatrixf(projectionTransformation.Get());
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadMatrixf(viewingTransformation.Get());
		glMultMatrixf(dataTransformation.Get());
		float mat1[16];
		glGetFloatv(GL_PROJECTION_MATRIX, mat1);
		projectionMatrix = mat1;
		float mat2[16];
		glGetFloatv(GL_MODELVIEW_MATRIX, mat2);
		modelMatrix = mat2;
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		const float zScale = GetPlugin().GetProperty("Z scale").require(Variant::TypeFloat());

		float width = volume->GetSize().GetX();
		float height = volume->GetSize().GetY();
		float depth = volume->GetSize().GetZ();

		// get x and y at center of volume
		Matrix translationCenterMatrix = Matrix(
			1, 0, 0, width / 2,
			0, 1, 0, height / 2,
			0, 0, 1, depth / 2,
			0, 0, 0, 1);
		Matrix matrixCenter = projectionMatrix * modelMatrix * translationCenterMatrix;
		float xCenter = matrixCenter.Get(12) / matrixCenter.Get(15);
		float yCenter = matrixCenter.Get(13) / matrixCenter.Get(15);

		// get largest distance from center to all eight points in volume cube
		float maxLength = 0;
		for (int i=0; i<8; i++) {
			Matrix translationCornerMatrix = Matrix(
				1, 0, 0, (i % 2) * width,
				0, 1, 0, ((int)(i / 2) % 2) * height,
				0, 0, 1, ((int)(i / 4) % 2) * depth * zScale,
				0, 0, 0, 1);
			Matrix matrixCorner = projectionMatrix * modelMatrix * translationCornerMatrix;
			float xCorner = matrixCorner.Get(12) / matrixCorner.Get(15);
			float yCorner = matrixCorner.Get(13) / matrixCorner.Get(15);
			float length = sqrt(pow(xCenter - xCorner, 2) + pow(yCenter - yCorner, 2));
			if (length > maxLength)
				maxLength = length;
		}

		// make points around cube for text boxes
		float radius = maxLength;
		for (unsigned int i=0; i<boxes.size(); i++) {
			float theta = (2 * boxes[i].depth + 3) * PI / 4;
			float rx = radius * cos(theta);
			float ry = radius * sin(theta);
			boxes[i].x = xCenter + rx + rx * (boxes[i].width / 2) / radius;
			boxes[i].y = yCenter + ry + ry * (boxes[i].height / 2) / radius;
		}
	}
}

/*
	test if position is inside a text box
*/
bool RendererLabel::hitTest(const Vector position, struct box *rect)
{
	if (position.GetX() < rect->x - rect->width / 2 - boxMargin)
		return false;
	if (position.GetX() > rect->x + rect->width / 2 + boxMargin)
		return false;
	if (position.GetY() > rect->y + rect->height / 2 + boxMargin)
		return false;
	if (position.GetY() < rect->y - rect->height / 2 - boxMargin)
		return false;
	return true;
}

void RendererLabel::reshape(const unsigned int width, const unsigned int height) {
	_width = width;
	_height = height;
}

void RendererLabel::updated(const Variant & varVariant, const Observable::Event & eveEvent)
{
	if (bool(GetPlugin().GetProperty("Enabled")))
		GetPlugin().update();
}

std::string DICOMItem::getValue()
{
	std::stringstream ss;
	ss << group << ": " << element;
	return ss.str();
}

DEFINE_RENDERERPLUGIN(RendererLabel,"Renderer Label Version 1.0",RendererLabel_png)
